
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <X11/Xlib.h>
#include "input.h"
#include "pnginput.h"
#include "xoutput.h"
#include "tools.h"
#include "leetft.h"
#include "speedy.h"
#include "rtctimer.h"
#include "vidmode.h"

/**
 * Set this to 1 to see lots of debug messages.
 */
static int verbose = 0;

/**
 * Pixels to move per second.
 */
static int pixels_per_second = 200;

/**
 * Set this to 1 if the flat panel you're using isn't giving reasonable
 * modeline information.
 */
static int believe_measured_rate = 0;


static int timediff( struct timeval *large, struct timeval *small )
{
    return (   ( ( large->tv_sec * 1000 * 1000 ) + large->tv_usec )
             - ( ( small->tv_sec * 1000 * 1000 ) + small->tv_usec ) );
}

static double intensity_to_voltage_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    /* sRGB uses an exponent of 2.4 in its nonstandard curve, but it
     *      * still advertises a display gamma of 2.2.
     *           */
    if( val <= 0.0031308 ) {
        return 12.92 * val;
    }
    return ( ( 1.055 * pow( val, 1.0 / 2.4 ) ) - 0.055 );
}

static int crawlpage = 0;
static int crawlpos = 0;
static pnginput_t *curpng = 0;

static int load_next_png( void )
{
    char filename[ 1024 ];

    if( curpng ) pnginput_delete( curpng );
    sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
    curpng = pnginput_new( filename );
    if( !curpng ) {
        crawlpage = 0;
        sprintf( filename, "/tmp/crawl%08d.png", crawlpage++ );
        curpng = pnginput_new( filename );
        if( !curpng ) {
            fprintf( stderr, "viewcrawl: Can't find %s: nothing to draw.\n", filename );
            return 0;
        }
    }
    crawlpos = 0;

    return 1;
}

static int fill_next_page( unsigned char *data, int width, int height, int stride )
{
    int curpos = 0;
    int startpos = 0;
    int i;
    int done = 0;

    while( curpos < width ) {
        int curwidth = pnginput_get_width( curpng ) - crawlpos;
        int loadnext = 1;

        startpos = crawlpos;
        if( curwidth >= ( width - curpos ) ) {
            curwidth = ( width - curpos );
            crawlpos += curwidth;
            loadnext = 0;
            done = 1;
        }

        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = data + (i * stride) + (curpos * 4);

            if( i >= pnginput_get_height( curpng ) ) {
                blit_colour_packed4444_scanline( curscanline, curwidth, 0, 0, 0, 0 );
            } else {
                rgba32_to_argb32_scanline( curscanline,
                                           pnginput_get_scanline( curpng, i ) + (startpos * 4),
                                           curwidth );
            }
        }
        curpos += curwidth;

        if( loadnext ) {
            if( !load_next_png() ) {
                return 0;
            }
        }
    }
    return 1;
}

int main( int argc, char **argv )
{
    output_api_t *output = xoutput_get_output();
    input_t *in = input_new();
    int width = 800;
    int height, stride;
    unsigned char *page1;
    unsigned char *page2;
    struct timeval lastblit;
    vidmode_t *vidmode;
    int pos = 0;
    int pagewidth, pageheight, pagestride;
    int usecperframe;
    int refreshtime;
    int accel_step;
    int accel_max;
    int accel = 0;
    int paused = 1;
    int framecount = 0;
    int prio = -19;
    int i;
    char c;

    /* Soon to support X-style -geometry, need this for tvtime also.
    int x = 0,y = 0,w = 0,h = 0;
    XParseGeometry( argv[ 1 ], &x, &y, &w, &h );
    fprintf( stderr, "%d,%d %d,%d\n", x, y, w, h );
    return 0;
    */

    while( (c = getopt( argc, argv, "vbs:p:w:" )) != -1 ) {
        switch( c ) {
        case 'v': verbose = 1; break;
        case 'p': pixels_per_second = atoi( optarg ); break;
        case 'b': believe_measured_rate = 1; break;
        case 'w': width = atoi( optarg ); break;
        case 'r': prio = atoi( optarg ); break;
        default:
            return 0;
        }
    }

    setup_speedy_calls( 1 );

    if( !load_next_png() ) {
        return 1;
    }

    if( setpriority( PRIO_PROCESS, 0, prio ) < 0 ) {
        fprintf( stderr, "viewcrawl: Can't renice to %d.\n", prio );
    }

    if( !set_realtime_priority( 0 ) ) {
        fprintf( stderr, "viewcrawl: Can't set realtime priority, must be root.\n" );
        fprintf( stderr, "viewcrawl: Accuracy will suffer.\n" );
    }

    fprintf( stderr, "viewcraw: Window width %d.\n", width );
    height = 40;
    if( !output->init( width, height ) ) {
        fprintf( stderr, "viewcrawl: No output available, exiting.\n" );
        return 1;
    }
    stride = output->get_output_stride();

    pagewidth = width;
    pageheight = height;
    pagestride = pagewidth * 4;

    page1 = (unsigned char *) malloc( pagestride * height );
    page2 = (unsigned char *) malloc( pagestride * height );
    if( !page1 || !page2 ) {
        fprintf( stderr, "viewcrawl: Can't allocte memory.\n" );
        return 1;
    }

    if( !fill_next_page( page1, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }
    if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
        fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
    }

    /* Try to measure refresh rate. */
    vidmode = vidmode_new();
    if( vidmode ) {
        double refresh = vidmode_get_refresh( vidmode );
        fprintf( stderr, "viewcrawl: Modeline refresh rate is %.2fhz.\n", refresh );
        refreshtime = (int) (((1.0 / refresh) * 1000.0 * 1000.0) + 0.5);
    } else {
        refreshtime = (int) (((1.0 / 85.0) * 1000.0 * 1000.0) + 0.5);
    }

    /* Blit a frame every refresh time. */
    usecperframe = refreshtime;

    /* Reset our start time. */
    fprintf( stderr, "viewcrawl: VidMode says %d usec per frame.\n", usecperframe );

    /* Test refresh rate. */
    {
        struct timeval bef, af;
        int sum = 0;
        int trials = 20;
        output->wait_for_retrace();
        output->wait_for_retrace();
        for( i = 0; i < trials; i++ ) {
            gettimeofday( &bef, 0 );
            output->wait_for_retrace();
            gettimeofday( &af, 0 );
            sum += timediff( &af, &bef );
        }
        fprintf( stderr, "viewcrawl: Roughly measured at %d usec over %d trials.\n",
                 sum / trials, trials );
        if( believe_measured_rate ) {
            fprintf( stderr, "viewcrawl: Assuming measured value is correct.\n" );
            usecperframe = sum / trials;
        }
        if( sum / trials < 5000 ) {
            fprintf( stderr, "viewcrawl: Either your refresh rate is >= 200hz, or something\n"
                             "viewcrawl: is wrong with your DRI setup.  Exiting to avoid problems.\n" );
            return 1;
        }
    }

    /* Start with 200 pixels per second. */
    fprintf( stderr, "viewcrawl: Speed is %d pixels per second.\n", pixels_per_second );
    accel_max = ((((pixels_per_second * usecperframe) / 1000))*65536)/1000;
    accel_step = accel_max / 8;

    output->wait_for_retrace();
    gettimeofday( &lastblit, 0 );
    framecount = 0;

    for(;;) {
        struct timeval beforewait;
        struct timeval beforeblit;
        struct timeval afterblit;

        /* Check for input. */
        output->poll_events( in );
        if( input_quit_requested( in ) ) break;
        if( input_pause_requested( in ) ) paused = !paused;
        input_next_frame( in );

        /* Update position. */
        if( !paused ) {
            pos += accel;
            if( accel < accel_max ) {
                accel += accel_step;
            }
        } else {
            pos += accel;
            if( accel > 0 ) {
                accel -= accel_step;
            }
        }

        /* Swap pages if necessary. */
        if( pos > (pagewidth << 16) ) {
            unsigned char *tmpp = page1;
            page1 = page2;
            page2 = tmpp;
            if( !fill_next_page( page2, pagewidth, pageheight, pagestride ) ) {
                fprintf( stderr, "viewcrawl: Can't open next crawl file, exiting.\n" );
                return 1;
            }
            pos -= (pagewidth << 16 );
        }

        /* Render image at the current position. */
        output->lock_output_buffer();
        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = output->get_output_buffer() + (stride * i);
            unsigned char *curpage1 = page1 + (i * pagestride);
            unsigned char *curpage2 = page2 + (i * pagestride);
            int pixpos = (pos >> 16);
            rgb32_subpix_blit_scanline( curscanline,
                                        curpage1 + (pixpos * 4),
                                        0, 0, 0,
                                        pos & 0xffff,
                                        pagewidth - pixpos );
            if( pos ) {
                unsigned char lastr = curpage1[ (pagewidth*4) - 2 ];
                unsigned char lastg = curpage1[ (pagewidth*4) - 3 ];
                unsigned char lastb = curpage1[ (pagewidth*4) - 4 ];
                rgb32_subpix_blit_scanline( curscanline + ((width - pixpos) * 4),
                                            curpage2, lastr, lastg, lastb,
                                            (pos & 0xffff), pixpos );
            }
        }
        output->unlock_output_buffer();

        /* Wait for the retrace before blitting. */
        gettimeofday( &beforewait, 0 );
        output->wait_for_retrace();

        /* Blit the frame. */
        gettimeofday( &beforeblit, 0 );
        output->show_frame();
        gettimeofday( &afterblit, 0 );

        /* Report frame drops. */
        if( timediff( &afterblit, &lastblit ) > ((usecperframe*2)+(usecperframe/2)) ) {
            fprintf( stderr, "viewcrawl: Frame dropped?\n" );
        }

        if( verbose ) {
            fprintf( stderr, "viewcrawl: %.1fms between blits, draw: %.1fms, blit: %.1fms.  Using %.1f%% of the CPU.\n",
                     ((double)timediff( &afterblit, &lastblit )) / 1000.0,
                     ((double)timediff( &beforewait, &lastblit )) / 1000.0,
                     ((double)timediff( &afterblit, &beforeblit )) / 1000.0,
                     100.0 * ((double) (timediff( &beforewait, &lastblit ) + timediff( &afterblit, &beforeblit ))) / ((double) timediff( &afterblit, &lastblit )) );
        }

        lastblit = afterblit;
        framecount++;
    }

    output->shutdown();
    return 0;
}

